;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; MASM Stack Probing PROLOGUE macro
; by chep, 2005/06/22
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; Allows a procedure to safely use LOCAL variables with a total size of 4kb or more,
; using an unrolled stack probing method.
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; Usage:
;
;   OPTION PROLOGUE:STACKPROBE
;   MyProcedure PROC ; ...
;     ; ...
;   MyProcedure ENDP
;   OPTION PROLOGUE:PROLOGUEDEF
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; Notes:
;   - When the total size of the LOCAL variables is less than 4kb, the code generated is
;     identical to PROLOGUEDEF, so there is no drawback using this macro
;   - See "OPTION PROLOGUE" and "PROC" topics in MASM32.HLP for the macro specifications
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; Limitations compared to PROLOGUEDEF:
;   - Stack probing is relevant only for Windows, ie FLAT model, so it won't accept other models
;   - Due to the FLAT model restriction, LOADDS is not supported
;   - FORCEFRAME argument doesn't generate a correct epilogue when no LOCAL variables are defined
;     So it is not supported for now :(
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; TODO:
;   - Allow looped probing additionally to the current unrolled probing
;     This behaviour should be controlled through macro arguments
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

STACKPROBE MACRO procname, flags, argbytes, localbytes, reglist, userparms:VARARG
  LOCAL parm, reg, probe, max_probe, line, page_size
  ;; Memory page size, can be changed for other architectures
  page_size = 4096
  ;; Throw an error if not in FLAT model, because stack probing is irrelevant otherwise
  IF @Model NE 7
    line TEXTEQU %@Line
    % ECHO @FileCur(line) : STACKPROBE prologue ERROR: stack probing is irrelevant if not using FLAT model
    .ERR
    EXITM <1>
  ENDIF
  ;; Detect macro arguments to give a warning
  IFNB <userparms>
    FOR parm,<userparms>
      line TEXTEQU %@Line
      % ECHO @FileCur(line) : STACKPROBE prologue warning: unknown prologue argument : parm
    ENDM
  ENDIF

  ;; Set up stack frame
  IF localbytes GT 0
    push ebp
    mov  ebp, esp
    IF localbytes LT page_size
      ;; Normal stack frame, no probing
      add  esp, (NOT localbytes)+1
    ELSE
      ;; Unrolled stack probing
      max_probe = (localbytes+page_size-1) AND (-page_size) ;; round up to next page size
      probe = page_size
      WHILE probe LE max_probe
        ;; Probe stack
        mov  DWORD PTR [ebp-probe], eax
        probe = probe + page_size
      ENDM
      add  esp, (-localbytes)
    ENDIF
  ELSEIF argbytes GT 0
    push ebp
    mov  ebp, esp
  ENDIF

  ;; USES clause
  IFNB <reglist>
    FOR reg,reglist
      push reg
    ENDM
  ENDIF

  EXITM <0>
ENDM
